#include "RACE_Logger.h"
#include "RACE_CmdDumpable.h"
#include "RACE_LogSwitchingStrategy.h"

envdebug *RACE_Logger::ms_debug = new envdebug("RACE_Logger");

/*------------- The Logger itself ----------------*/

RACE_Logger::RACE_Logger(RACE_Logfile *logfile,const char *name)
    : RACE_Synch_Task(&m_qls,0),
      mp_logfile(logfile),
      m_accu(logfile),
      mp_LogSwitchingStrategy(new RACE_DefaultLogSwitchingStrategy(this)),
      m_canceled(false),
      m_qls(name),
      m_refcount(1)
{}

const char *RACE_Logger::GetName()
{
  return m_qls.GetQueueName();
}

void RACE_Logger::Log(RACE_Dumpable *dumpable)
{
  /*
   * During the Canceling-Operation in the Flush-Operation the Logging of
   * Snapshots may be triggered - this happens in the Loggerthread
   */
  if(m_canceled && dumpable->length() > 0 && !m_qls.IsCallerLoggerThread())
  {
    output::warn("RACE_Logger::log: Logger canceled - discarding dumpable");
    dumpable->Discard();
    dumpable->release();
    return;
  }
  int r=msg_queue()->enqueue(dumpable);
  ms_debug->print("RACE_Logger::log enqueue returns %d",r);
  if(r<0 && errno==EWOULDBLOCK)
  {
    output::error("EnQueuing TIMEOUT - Canceling");
    Cancel();
    // Cancel Imediately,
    // don't wait until the Canceldumpable reaches the logger
    m_canceled=true;
    return;
  }
  output::cond_fail(r<0,"Logger::log enqueue");
}

int RACE_Logger::svc(void)
{
  int count=1;
  RACE_Message_Block *mb;
  m_qls.SetLoggerThreadId(ACE_OS::thr_self());
  for(;!(m_canceled && msg_queue()->is_empty()); count++)
  {
    ms_debug->print("(%t) in iteration %d before getq()\n",count);
    if (this->getq(mb) == -1)
    {
      m_accu.Flush(); // If error, we want to have all on disk
      output::error("(%t) in iteration %d, got result -1, exiting\n",count);
      break;
    }
    RACE_Dumpable *d=ACE_dynamic_cast(RACE_Dumpable *,mb);
    int length = d->total_length();

    if (length > 0)
    {
      ms_debug->print("(%t) in iteration %d, length = %d fragmented=%s\n",
                   count,length,(d->IsFragmented() ? "true" : "false"));
      if(d->IsDiscardable() && m_qls.is_critical_watermark_reached())
      {
        output::warn("RACE_Logger: Discarding dispensable Dumpable");
        d->Discard();
      }
      else
      {
        m_accu.Add(d);
        mp_LogSwitchingStrategy->InfoCurrentOffset(m_accu.GetCurrentOffset());
      }
    }
    else
    {
      if(ACE_dynamic_cast(RACE_CancelCmdDumpable *,d))
      {
        ms_debug->print("RACE_Logger: RACE_CancelCmdDumpable");
        m_accu.Flush(); // some Dumpables may produced here
        m_canceled = true; // dissallow queueing
      }
      else if(ACE_dynamic_cast(RACE_FlushCmdDumpable *,d))
      {
        ms_debug->print("RACE_Logger: RACE_FlushCmdDumpable");
        m_accu.Flush();
      }
      else
      {
        output::error("RACE_Logger received an Dumpable with length=%d, queuelength=",length,msg_queue()->message_count());
      }
      d->release();
    }
  }
  m_accu.Flush(); // Flush the last Dumpables
  logger_thread_stopped();
  return 0;
}



void RACE_Logger::Cancel()
{
  RACE_Dumpable *d = new RACE_CancelCmdDumpable();
  /*
   * In all cases, the CancesDumbable must be enqueued
   * The CActiveObject must be possible to flush its queue and so on
   */
  int r=msg_queue()->enqueue_force(d);
  ms_debug->print("RACE_Logger::Cancel: enqueue_force returns %d",r);
  output::cond_fail(r<0,"CActiveConsumer::enqueue_force");
}

void RACE_Logger::Flush()
{
  RACE_Dumpable *d = new RACE_FlushCmdDumpable();
  Log(d);
}

RACE_Logger* RACE_Logger::Duplicate()
{
  m_refcount++;
  ms_debug->print("RACE_Logger::duplicate: %d References",m_refcount);
  return this;
}

void RACE_Logger::Release()
{
  ms_debug->print("RACE_Logger::release: %d ",m_refcount);
  m_refcount--;
  if(m_refcount == 0)
  {
    ms_debug->print("RACE_Logger::release: would delete myself: thr_count=%d",
                 thr_count());
  }
}

void RACE_Logger::SwitchLog()
{
  m_accu.LogSwitch();
  mp_logfile->SwitchPart();
}

void RACE_Logger::InfoSnapshotWritten()
{
  mp_LogSwitchingStrategy->InfoSnapshotWritten();
}

void RACE_Logger::logger_thread_stopped()
{
  ms_debug->print("RACE_Logger::logger_thread_stopped() called");
}

RACE_Logger::~RACE_Logger()
{
  ms_debug->print("RACE_Logger::~RACE_Logger()");
}


#if defined (ACE_HAS_EXPLICIT_TEMPLATE_INSTANTIATION)
template class RACE_Message_Queue<ACE_MT_SYNCH>;
template class RACE_Task<ACE_MT_SYNCH>;
#elif defined (ACE_HAS_TEMPLATE_INSTANTIATION_PRAGMA)
#pragma instantiate RACE_Message_Queue<ACE_MT_SYNCH>;
#pragma instantiate RACE_Task<ACE_MT_SYNCH>;
#endif
