/**
 *  @file    RACE_Message_Queue_T.h
 *  counterfeit of ace/Message_Queue_T.h
 */
#ifndef RACE_MESSAGE_QUEUE_T_H
#define RACE_MESSAGE_QUEUE_T_H
#include <ace/pre.h>
#include <ace/Synch.h>
#include "RACE_Message_Queue.h"
#include "RACE_QueueLevelStrat.h"

#if !defined (ACE_LACKS_PRAGMA_ONCE)
# pragma once
#endif /* ACE_LACKS_PRAGMA_ONCE */

//##ModelId=3F72A5C501E7
//##Documentation
//##  @class RACE_Message_Queue
//##
//##  @brief A threaded message queueing facility, modeled after the
//##  queueing facilities in System V STREAMs.
//##
//##  An <RACE_Message_Queue> is the central queueing facility for
//##  messages in the ACE framework.  If <ACE_SYNCH_DECL> is
//##  <ACE_MT_SYNCH> then all operations are thread-safe.
//##  Otherwise, if it's <ACE_NULL_SYNCH> then there's no locking
//##  overhead.

template <ACE_SYNCH_DECL>
class RACE_Message_Queue : public RACE_Message_Queue_Base
{
 public:
  // = Initialization and termination methods.
//##ModelId=3F72A5CF00CA
//##Documentation
//##  Initialize an <RACE_Message_Queue>.  The <high_water_mark>
//##  determines how many bytes can be stored in a queue before it's
//##  considered "full."  Supplier threads must block until the queue
//##  is no longer full.  The <low_water_mark> determines how many
//##  bytes must be in the queue before supplier threads are allowed to
//##  enqueue additional <RACE_Message_Block>s.  By default, the
//##  <high_water_mark> equals the <low_water_mark>, which means that
//##  suppliers will be able to enqueue new messages as soon as a
//##  consumer removes any message from the queue.  Making the
//##  <low_water_mark> smaller than the <high_water_mark> forces
//##  consumers to drain more messages from the queue before suppliers
//##  can enqueue new messages, which can minimize the "silly window
//##  syndrome."

  RACE_Message_Queue(RACE_QueueLevelStrategy *qls);

//##ModelId=3F72A5CF00D1
//##Documentation
//##  Initialize an <RACE_Message_Queue>.
  virtual int open();

//##ModelId=3F72A5CF00D4
//##Documentation
//## Close down the message queue and release all resources.
  virtual int close(void);

//##ModelId=3F72A5CF00DC
//##Documentation
//## Close down the message queue and release all resources.
  virtual ~RACE_Message_Queue (void);

    // = Enqueue and dequeue methods.

  // For the following enqueue and dequeue methods if <timeout> == 0,
  // the caller will block until action is possible, else will wait
  // until the absolute time specified in *<timeout> elapses).  These
  // calls will return, however, when queue is closed, deactivated,
  // when a signal occurs, or if the time specified in timeout
  // elapses, (in which case errno = EWOULDBLOCK).

  /**
   * Retrieve the first <RACE_Message_Block> without removing it.  Note
   * that <timeout> uses <{absolute}> time rather than <{relative}>
   * time.  If the <timeout> elapses without receiving a message -1 is
   * returned and <errno> is set to <EWOULDBLOCK>.  If the queue is
   * deactivated -1 is returned and <errno> is set to <ESHUTDOWN>.
   * Otherwise, returns -1 on failure, else the number of items still
   * on the queue.
   */
  //virtual int peek_dequeue_head (RACE_Message_Block *&first_item, ACE_Time_Value *timeout = 0);

//##ModelId=3F72A5CF00E3
//##Documentation
//##  Enqueue an <RACE_Message_Block *> into the <Message_Queue>
//##  FIFO Order is maintained
//##  The Timeouts are set via the RACE_QueuelevelStrategy
//##  If the <timeout> elapses
//##  without receiving a message -1 is returned and <errno> is set to
//##  <EWOULDBLOCK>.  If the queue is deactivated -1 is returned and
//##  <errno> is set to <ESHUTDOWN>.  Otherwise, returns -1 on failure,
//##  else the number of items still on the queue.

  virtual int enqueue (RACE_Message_Block *new_item);

    //##ModelId=3F72A5CF00EB
  //##Documentation
  //##  Enqueue an <RACE_Message_Block *> into the <Message_Queue>
  //##  without regarding the full-Condition
  //##  FIFO Order is maintained
  //##  This is thought only for enqueing Finition-Messages
  //##  which must be enqueued in the case when the Queue is full
  int enqueue_force(RACE_Message_Block *new_item);

//##ModelId=3F72A5CF00F3
//##Documentation
//##  Dequeue and return the <ACE_Message_Block *> at the head of the
//##  queue.  Note that <timeout> uses <{absolute}> time rather than
//##  <{relative}> time.  If the <timeout> elapses without receiving a
//##  message -1 is returned and <errno> is set to <EWOULDBLOCK>.  If
//##  the queue is deactivated -1 is returned and <errno> is set to
//##  <ESHUTDOWN>.  Otherwise, returns -1 on failure, else the number
//##  of items still on the queue.

  virtual int dequeue(RACE_Message_Block *&first_item);

//##ModelId=3F72A5CF00FC
//##Documentation
//## Returns a reference to the lock used by the <ACE_Message_Queue>.
  ACE_SYNCH_MUTEX_T &lock (void)
    {
      // The Sun Forte 6 (CC 5.1) compiler is only happy if this is in the
      // header file      (j.russell.noseworthy@objectsciences.com)
      return this->lock_;
    }

    //##ModelId=3F72A5CF0102
  int is_empty(void);
    //##ModelId=3F72A5CF0109
  int is_full(void);

   // = Queue statistic methods.
//##ModelId=3F72A5CF0110
//##Documentation
//##  Number of total bytes on the queue, i.e., sum of the message
//##  block sizes.
//##  Number of total length on the queue, i.e., sum of the message
//##  block lengths.
//##  Number of total messages on the queue.

  virtual size_t message_bytes (void);
    //##ModelId=3F72A5CF0119
  virtual size_t message_length (void);
    //##ModelId=3F72A5CF0120
  virtual size_t message_count (void);
protected:
  // = Synchronization primitives for controlling concurrent access.
//##ModelId=3F72A5CF009B
//##Documentation
//## Protect queue from concurrent access.
  ACE_SYNCH_MUTEX_T lock_;
//##ModelId=3F72A5CF009E
//##Documentation
//## Used to make threads sleep until the queue is no longer empty.
  ACE_SYNCH_CONDITION_T not_empty_cond_;
//##ModelId=3F72A5CF00A0
//##Documentation
//## Used to make threads sleep until the queue is no longer full.
  ACE_SYNCH_CONDITION_T not_full_cond_;
  // = Routines that actually do the enqueueing and dequeueing.

  // These routines assume that locks are held by the corresponding
  // public methods.  Since they are virtual, you can change the
  // queueing mechanism by subclassing from <RACE_Message_Queue>.

//##ModelId=3F72A5CF0127
//##Documentation
//## Enqueue an <RACE_Message_Block *>
  virtual int enqueue_tail_i(RACE_Message_Block *new_item);
    //##ModelId=3F72A5CF0130
  virtual int dequeue_head_i(RACE_Message_Block *&first_item);

  // = Helper methods to factor out common #ifdef code.

//##ModelId=3F72A5CF0139
//##Documentation
//## Wait for the queue to become non-full.
  virtual int wait_not_full_cond(ACE_Time_Value *timeout);

//##ModelId=3F72A5CF0141
//##Documentation
//## Wait for the queue to become non-empty.
  virtual int wait_not_empty_cond(ACE_Time_Value *timeout);

//##ModelId=3F72A5CF014B
//##Documentation
//## Inform any threads waiting to enqueue that they can procede.
  virtual int signal_enqueue_waiters(void);

//##ModelId=3F72A5CF0153
//##Documentation
//## Inform any threads waiting to dequeue that they can procede.
  virtual int signal_dequeue_waiters(void);

//##ModelId=3F72A5CF00A4
//##Documentation
//## Pointer to head of RACE_Message_Block list.
  RACE_Message_Block *head_;

//##ModelId=3F72A5CF00B1
//##Documentation
//## Pointer to tail of RACE_Message_Block list.
  RACE_Message_Block *tail_;


    //##ModelId=3F72A5CF00BE
  RACE_QueueLevelStrategy *qls_;

 private:
  // = Disallow these operations.
  ACE_UNIMPLEMENTED_FUNC (void operator= (const RACE_Message_Queue<ACE_SYNCH_USE> &))
  ACE_UNIMPLEMENTED_FUNC (RACE_Message_Queue (const RACE_Message_Queue<ACE_SYNCH_USE> &))
};

#if defined (ACE_TEMPLATES_REQUIRE_SOURCE)
#include "RACE_Message_Queue_T.cpp"
#endif /* ACE_TEMPLATES_REQUIRE_SOURCE */

#if defined (ACE_TEMPLATES_REQUIRE_PRAGMA)
#pragma implementation ("RACE_Message_Queue_T.cpp")
#endif /* ACE_TEMPLATES_REQUIRE_PRAGMA */

#include <ace/post.h>
#endif /* RACE_MESSAGE_QUEUE_T_H */
