#include "output.h"
#include "env_access.h"

#include <cstdio>
#include <ctime>
#include <iostream>
#include <cstdlib>
#include <unistd.h>
#include <csignal>
#include <string>
#include <syslog.h>
#include <sys/timeb.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/stat.h>

/*
 * $Id:  $
 */

namespace output {

// Implementation.

// #######################################################################################
// private in namespace:
// #######################################################################################

component component::the_default_component ("");

struct sembuf output_sem_lock[2] = { { 0, 0, 0 }, { 0, 1, SEM_UNDO } };
struct sembuf output_sem_unlock[1] = { { 0, -1, SEM_UNDO } };

std::string executable_name;
std::string language ("C++");

bool time_output = false;

int semaphore_id = semget (8476572, 1, IPC_CREAT | S_IRUSR | S_IWUSR);
bool semaphore_usage_allowed = false;
bool semaphore_usage_uninitialized = true;

bool syslog_usage_allowed = false;
bool syslog_usage_uninitialized = true;

#define var_arg_to_string(METHODNAME, XXX) \
    if (XXX == NULL) { return; } \
    va_list vargu; \
    char str[4096]; \
    va_start (vargu, XXX); \
    vsnprintf (str, 4096, XXX, vargu); \
    va_end (vargu)

void set_output_executable_name ()
{
    if (executable_name.size () > 0)
        return; // the name was already set

    char command[256];

    // try to read file /proc/<pid>/cmdline first
    sprintf (command, "/proc/%i/cmdline", getpid ());
    FILE *cmdline_file_ptr = fopen (command, "r");
    if (cmdline_file_ptr != NULL) {
      size_t read_bytes = fread (command, 1, 255, cmdline_file_ptr);
      fclose (cmdline_file_ptr);
      if (read_bytes  > 0) {
        command[255] = '\0';
        int len = strlen (command);
        while (len >= 0) {
            if (command[len] == '/')
                break;
            len--;
        }
        executable_name = &command[len + 1];
        return;
      }
    }

    // reading /proc/<pid>/cmdline failed ==> try to use popen...

    // may be the following ps call is not the most elegant way but it
    // should work on many different systems
    // alternatives are:
    //   "ps o pid,command | awk '{if ($1+0 == %i) print($2); }'", getpid ()    <--- OSF1
    //   "ps chp %i | awk '{print $(NF);}'", getpid ()                          <--- LINUX

    sprintf (command, "ps p %i | grep %i | awk '{if ($4 != \"+\")  print($5); else print($6); }'", getpid (), getpid ());

    FILE *fp = popen (command, "r");
    if (fp == NULL)
        executable_name = "UnknownExecutable";
    else {
        char helparray[256];
        if (fgets (helparray, 256, fp) == NULL)
            executable_name = "UnknownExecutable";
        else {
            int len = strlen (helparray);
            if (helparray[len = (len ? len - 1 : 0)] == '\n')
                helparray[len] = 0; // the awk print writes a newline!!!
            while (len >= 0) {
                if (helparray[len] == '/')
                    break;
                len--;
            }
            executable_name = &helparray[len + 1];
        }
        pclose (fp);
    }
    return;
}

void check_for_syslog_usage ()
{
    // check for syslog usage
    if (syslog_usage_uninitialized) {
        env_flag ef_syslog_usage_allowed ("UTILITY_USE_SYSLOG");
        syslog_usage_allowed = ef_syslog_usage_allowed.is_on ();
        syslog_usage_uninitialized = false;
    }
}


// #######################################################################################
// public:
// #######################################################################################

void component::general_output_function (output_classification kindof, char *str)
{
    // generate an output of the form: <executable_name>:: [<component_name>] <message_type>: <message>
    char prefix_output_string[1024];
    char time_information[10] = { 0, 0, 0, 0, 0, 0, 0, 0, ' ', 0 };

    if (str == NULL) {
        general_output_function (e_warn, "output::component::general_output_function parameter error");
        return;
    }

    prefix_output_string[0] = '\0';

    check_for_syslog_usage ();

    if (! syslog_usage_allowed) {
        set_output_executable_name ();
        strcat (prefix_output_string, executable_name.c_str ());
        strcat (strcat (strcat (prefix_output_string, "["), language.c_str ()), "]:: ");
        if (time_output) {
            time_t tt_var;
            if (time (&tt_var) < 0)
                general_output_function (e_warn, "output::component::general_output_function - time call failed");
            else
                strncpy (time_information, ctime (&tt_var)+11, 8);
        }
    }

    std::string usertext = ""; // initialization is important here because otherwise the buggy STL
                               // implementation of gcc 2.95.3 may crash in usertext.append (str)
    if (m_name.size () > 0) {
        usertext = m_name;
        usertext.append (": ");
    }
    usertext.append (str);
    const char *ctext = usertext.c_str();
    // ctext is required because
    // "cannot pass objects of non-POD type `std::string' through `...'"

    // a semaphore might be used to avoid overlapping outputs
    if (semaphore_usage_uninitialized) {
        env_flag ef_semaphore_usage_allowed ("UTILITY_SERIALIZED_OUTPUT");
        semaphore_usage_allowed = ef_semaphore_usage_allowed.is_on ();
        semaphore_usage_uninitialized = false;
    }
    if ((semaphore_id != -1) && semaphore_usage_allowed)
        semop (semaphore_id, &output_sem_lock[0], 2);

    switch (kindof) {
        case e_warn       :
            if (syslog_usage_allowed)
                syslog (LOG_WARNING, "Warning: %s%s", prefix_output_string, ctext);
            else
                std::cerr << time_information << prefix_output_string << "Warning: " << ctext << std::endl;
            break;
        case e_error      :
            if (syslog_usage_allowed)
                syslog (LOG_ERR, "Error  : %s%s", prefix_output_string, ctext);
            else
                std::cerr << time_information << prefix_output_string << "Error  : " << ctext << std::endl;
            break;
        case e_fail       :
            if (syslog_usage_allowed)
                syslog (LOG_EMERG, "Failure: %s%s", prefix_output_string, ctext);
            else
                std::cerr << time_information << prefix_output_string << "Failure: " << ctext << std::endl;
            break;
        case e_unspecific :
            if (syslog_usage_allowed)
                syslog (LOG_NOTICE, "Notice : %s", ctext);
            else
                std::cerr << ctext << std::endl;
            break;
        case e_debug      :
            if (syslog_usage_allowed)
                syslog (LOG_DEBUG, "Debug  : %s%s", prefix_output_string, ctext);
            else
                std::cout << time_information << prefix_output_string << "Debug  : " << ctext << std::endl;
            break;
        case e_message    :
            if (syslog_usage_allowed)
                syslog (LOG_INFO, "Message: %s%s", prefix_output_string, ctext);
            else
                std::cout << time_information << prefix_output_string << "Message: " << ctext << std::endl;
            break;
    }
    if ((semaphore_id != -1) && semaphore_usage_allowed)
        semop (semaphore_id, &output_sem_unlock[0], 1);
}

void component::debug (char *msg, ...)
{
    var_arg_to_string (debug,msg);
    general_output_function (e_debug, str);
}

void component::message (char *msg, ...)
{
    var_arg_to_string (message,msg);
    general_output_function (e_message, str);
}

void component::warn (char *msg, ...)
{
    var_arg_to_string (warn,msg);
    general_output_function (e_warn, str);
}

void component::error (char *msg, ...)
{
    var_arg_to_string (error,msg);
    general_output_function (e_error, str);
}

void component::report (char *msg, ...)
{
    var_arg_to_string (report,msg);
    general_output_function (e_unspecific, str);
}

void component::fail (char *msg, ...)
{
    var_arg_to_string (fail,msg);
    general_output_function (e_fail, str);
#ifdef FAIL_PERMITS_TERM_HANDLER
    kill (getpid (), SIGTERM); // try a clean shutdown via a possible termination handler
    sleep (10); // give the handler a chance
#endif
    abort (); // terminating with SIGABRT and creating an analysable core file
    exit (EXIT_FAILURE);
}

void component::cond_debug (bool validity, char *msg, ...)
{
    if (validity) {
        var_arg_to_string (cond_debug,msg);
        general_output_function (e_debug, str);
    }
}

void component::cond_message (bool validity, char *msg, ...)
{
    if (validity) {
        var_arg_to_string (cond_message,msg);
        general_output_function (e_message, str);
    }
}

void component::cond_warn (bool validity, char *msg, ...)
{
    if (validity) {
        var_arg_to_string (cond_warn,msg);
        general_output_function (e_warn, str);
    }
}

void component::cond_error (bool validity, char *msg, ...)
{
    if (validity) {
        var_arg_to_string (cond_error,msg);
        general_output_function (e_error, str);
    }
}

void component::cond_fail (bool validity, char *msg, ...)
{
    if (validity) {
        var_arg_to_string (cond_fail,msg);
        general_output_function (e_fail, str);
#ifdef FAIL_PERMITS_TERM_HANDLER
        kill (getpid (), SIGTERM); // try a clean shutdown via a possible termination handler
        sleep (10); // give the handler a chance
#endif
        abort (); // terminating with SIGABRT and creating an analysable core file
        exit (EXIT_FAILURE);
    }
}

const char *hex_chars = "0123456789ABCDEF";

void component::dump_hex (const unsigned char *buffer, unsigned buffer_len, unsigned bytes_per_line)
{
    cond_fail (bytes_per_line < 1, "output::component::dump_hex: got wrong bytes_per_line value");

    unsigned output_bytes_per_line = bytes_per_line * 3;
    char hex_message[output_bytes_per_line];
    unsigned hex_chars_index = 0, buffer_offset = 0;
    u_char byte_var, *buffer_runner = (u_char *) buffer;

    while (buffer_offset < buffer_len) {
        byte_var = *buffer_runner;
        byte_var >>= 4; // process the upper 4 bits
        hex_message [hex_chars_index++] = hex_chars[byte_var];
        byte_var = *buffer_runner; buffer_runner++; buffer_offset++;
        byte_var &= 0xf; // process the lower 4 bits
        hex_message [hex_chars_index++] = hex_chars[byte_var];
        if (hex_chars_index >= output_bytes_per_line - 1) {
            hex_message [hex_chars_index] = 0;
            message ("%s", hex_message);
            hex_chars_index = 0;
        }
        else
            hex_message [hex_chars_index++] = ' ';
    }
    if (hex_chars_index > 0) {
        hex_message [hex_chars_index - 1] = 0;
        message ("%s", hex_message);
    }
}


component::component (const char *name)
{
    if (name == NULL)  // This is really a misuse.
        m_name = "";
    else
        m_name = name;
}

// #######################################################################################
// public in namespace:
// #######################################################################################

void set_alternative_logging_capability ()
{
    check_for_syslog_usage ();
    if (syslog_usage_allowed)
        openlog (NULL, 0, LOG_LOCAL0);
}

void set_language (const char *name)
{
    if (name)
        language = name;
}

void set_executable_name (const char *name)
{
    if (name == NULL)
        return; // handle invalid argument

    char *exename_temp = rindex (name, '/'); // skip path

    if ((exename_temp == NULL) || (strlen (exename_temp) <= 1))
        executable_name = name;
    else
        executable_name = exename_temp + 1;
}

const char *get_executable_name ()
{
    set_output_executable_name ();
    return strdup (executable_name.c_str ());
}

void time_output_on ()
{
    time_output = true;
}

void time_output_off ()
{
    time_output = false;
}

// For interfacing to other languages:
extern "C" void output_set_language (const char *name)
{
  output::set_language (name);
}


// Public in namespace:
// Non-object-bound functions

void debug (char *msg, ...)
{
    var_arg_to_string (debug,msg);
    component::get_default_component ().general_output_function (e_debug, str);
}

void message (char *msg, ...)
{
    var_arg_to_string (message,msg);
    component::get_default_component ().general_output_function (e_message, str);
}

void warn (char *msg, ...)
{
    var_arg_to_string (warn,msg);
    component::get_default_component ().general_output_function (e_warn, str);
}

void error (char *msg, ...)
{
    var_arg_to_string (error,msg);
    component::get_default_component ().general_output_function (e_error, str);
}

void report (char *msg, ...)
{
    var_arg_to_string (report,msg);
    component::get_default_component ().general_output_function (e_unspecific, str);
}

void fail (char *msg, ...)
{
    var_arg_to_string (fail,msg);
    component::get_default_component ().general_output_function (e_fail, str);
#ifdef FAIL_PERMITS_TERM_HANDLER
    kill (getpid (), SIGTERM); // try a clean shutdown via a possible termination handler
    sleep (10); // give the handler a chance
#endif
    abort (); // terminating with SIGABRT and creating an analysable core file
    exit (EXIT_FAILURE);
}

void cond_debug (bool validity, char *msg, ...)
{
    if (validity) {
        var_arg_to_string (cond_debug,msg);
        component::get_default_component ().general_output_function (e_debug, str);
    }
}

void cond_message (bool validity, char *msg, ...)
{
    if (validity) {
        var_arg_to_string (cond_message,msg);
        component::get_default_component ().general_output_function (e_message, str);
    }
}

void cond_warn (bool validity, char *msg, ...)
{
    if (validity) {
        var_arg_to_string (cond_warn,msg);
        component::get_default_component ().general_output_function (e_warn, str);
    }
}

void cond_error (bool validity, char *msg, ...)
{
    if (validity) {
        var_arg_to_string (cond_error,msg);
        component::get_default_component ().general_output_function (e_error, str);
    }
}

void cond_fail (bool validity, char *msg, ...)
{
    if (validity) {
        var_arg_to_string (cond_fail,msg);
        component::get_default_component ().general_output_function (e_fail, str);
#ifdef FAIL_PERMITS_TERM_HANDLER
        kill (getpid (), SIGTERM); // try a clean shutdown via a possible termination handler
        sleep (10); // give the handler a chance
#endif
        abort (); // terminating with SIGABRT and creating an analysable core file
        exit (EXIT_FAILURE);
    }
}

void dump_hex (const unsigned char *buffer, unsigned buffer_len, unsigned bytes_per_line)
{
    component::get_default_component ().dump_hex (buffer, buffer_len, bytes_per_line);
}

}  // end of namespace output

