#ifndef STRING_PARSER_H_
#define STRING_PARSER_H_

/*
 * $Id:$
 */

#include <string>
#include <ace/Vector_T.h>

//##Documentation
//## Base interface class for accessing the string to be parsed.
class string_parser_base {
  public:
    virtual std::string get_str () const = 0;
    virtual ~string_parser_base () {}
};

//##Documentation
//## Build list of strings by splitting given input.
class string_list_parser : public string_parser_base {
  public:

    //##Documentation
    //## Destructs an instance of string_list_parser.
    virtual ~string_list_parser () {}

    typedef ACE_Vector<std::string> StringVector;

    //##Documentation
    //## Split the delimiter separated list of values into its
    //## constituent values.
    //## Returns a vector of strings (the individual values)
    //## Example: for the input string ":Abc:Def" and delimiter
    //## ':', a three element vector is returned:
    //##  [0]  ""
    //##  [1]  "Abc"
    //##  [2]  "Def"
    StringVector splitValues (char delimiter = ':') const;

    //##Documentation
    //## Same as splitValues, but discard empty values (i.e. two subsequent
    //## colons, or a leading or trailing colon), and return an indication
    //## of whether empty values were encountered.
    //## Example: for the input string  ":Abc:Def" and delimiter
    //## ':', a two element vector is returned:
    //##  [0]  "Abc"
    //##  [1]  "Def"
    //## and the emptyValuesEncountered arg is returned as true (because of
    //## the leading colon.)
    StringVector splitAndPruneValues (bool& emptyValuesEncountered, char delimiter = ':') const;
};

//##Documentation
//## Check for switch state by parsing given input.
class string_flag_parser : public string_parser_base {
  public:

    //##Documentation
    //## Constructs an instance of string_flag_parser.
    string_flag_parser () : switch_state_checked (false), switch_is_on (false) {}

    //##Documentation
    //## Destructs an instance of string_flag_parser.
    virtual ~string_flag_parser () {}

    //##Documentation
    //## The parse method. Result is the state of the switch.
    bool is_on ();

  private:
    bool switch_state_checked;
    bool switch_is_on;
};

//##Documentation
//## Check the switch level by parsing given input.
class string_level_parser : public string_parser_base {
  public:

    //##Documentation
    //## Constructs an instance of string_level_parser.
    string_level_parser () : switch_level_checked (false), switch_level (0) {}

    //##Documentation
    //## Destructs an instance of string_level_parser.
    virtual ~string_level_parser () {}

    //##Documentation
    //## The parse method. Returns the level.
    unsigned long level ();

    //##Documentation
    //## Returns the state of the level.
    bool is_on ();

  private:
    bool switch_level_checked;
    unsigned long switch_level;
};

#endif

